const fs = require("fs");
const path = require("path");
const EventEmitter = require("events");
const { coding } = require("../util/coding");
const { decoding } = require("../util/decod");
const {
  VarietyOfIssues,
  AccessDenied,
  ExtensionMismatch,
  DataStructureIssue,
} = require("../util/errorHandling");

function getKeys(obj) {
  if (typeof(obj) === "object") return Object.keys(obj);
}

async function getData(tableName, directory, extName) {
  const filePath = path.join(process.cwd(), directory, tableName, "storage" + extName);
  try {
    let data = await fs.promises.readFile(filePath, "utf8");
    data = typeof data == "string" ? JSON.parse(data) : data;
    return data;
  } catch (error) {
    console.error("Error reading file:", error);
    return null;
  }
}

async function saveData(tableName, directory, db, extName, encryption = false) {
  const content = JSON.stringify(db);
  const filePath = path.join(process.cwd(), directory, tableName, "storage" + extName);
  try {
    await fs.promises.writeFile(filePath, content);
  } catch (error) {
    console.error("Error writing file:", error);
  }
}

class Storage extends EventEmitter {
  constructor(options) {
    super();
    this.directory = options && options.path ? options.path : "database";
    this.tables = options && Array.isArray(options.table) ? options.table : ["main"];
    this.extName = options && options.extname ? options.extname : ".sql";
    this.key = options && options.key ? options.key : '';

    if (!fs.existsSync(path.join(process.cwd(), this.directory))) {
      fs.promises.mkdir(path.join(process.cwd(), this.directory));
    }

    let tableList = this.tables;
    if (!Array.isArray(tableList)) {
      throw new VarietyOfIssues("Invalid table array");
    }
    if (!this.extName) {
      throw new ExtensionMismatch("Invalid extension");
    }

    for (let i = 0; i < tableList.length; i++) {
      const tablePath = path.join(process.cwd(), `${this.directory}/${tableList[i]}`);
      if (!fs.existsSync(tablePath)) {
        fs.promises.mkdir(tablePath);
      }

      const storagePath = path.join(process.cwd(), this.directory, tableList[i], `storage${this.extName}`);
      if (!fs.existsSync(storagePath)) {
        fs.promises.writeFile(storagePath, "{}");
      }
    }
  }

  async setData(tableName, key, value, encryption = false) {
    if (!this.tables.find((table) => table === tableName)) {
      throw new DataStructureIssue("Invalid table name: " + tableName);
    }
    const oldValue = await this.getData(tableName, key);
    let db = await getData(tableName, this.directory, this.extName);
    let values = value;
    if (encryption) {
      if (!this.key) {
        throw new AccessDenied("You need to specify 'key' in the class CreateStorage");
      }
      db[key] = {
        key: key,
        value: coding(values, this.key),
      };
    } else {
      db[key] = {
        key: key,
        value: values,
      };
    }
    const newValue = {
      table: tableName,
      path: this.directory,
      extname: this.extName,
      key: key,
      value: value,
      encryption: encryption,
    };
    await saveData(tableName, this.directory, db, this.extName);
    this.emit("update", newValue, oldValue);
  }

  async getData(tableName, key, encryption = "default") {
    if (!this.tables.find((table) => table === tableName)) {
      throw new DataStructureIssue("Invalid table name: " + tableName);
    }
    let db = await getData(tableName, this.directory, this.extName);
    if (await this.hasData(tableName, key)) {
      let values = db[key]["value"];
      if (encryption === true) {
        if (!this.key) {
          throw new AccessDenied("You need to specify 'key' in the class Robin");
        }
        return coding(values, this.key);
      } else if (encryption === false) {
        if (!this.key) {
          throw new AccessDenied("You need to specify 'key' in the class CreateStorage");
        }
        return decoding(values, this.key);
      } else if (encryption === "default") {
        return values;
      } else {
        throw new VarietyOfIssues("this type is invalid to 'get'");
      }
    } else {
      return undefined;
    }
  }

  async hasData(tableName, key) {
    let db = await getData(tableName, this.directory, this.extName);
    return db.hasOwnProperty(key);
  }
}

module.exports = { Storage };